import datetime
from base64 import b64decode
from collections import defaultdict
from dataclasses import InitVar, dataclass, field
from enum import Enum
from io import BytesIO
from typing import ClassVar, List, Optional, Union

import astropy.timeseries
import astropy.units as u
import astropy_healpix as ah
import numpy as np
import pandas as pd
from astropy.coordinates import SkyCoord
from astropy.table import Table
from typing_extensions import TypedDict

from ._api.api import _get_resource, _list_resources
from ._api.schemas import (
    _AlertSchema,
    _CatalogEntrySchema,
    _GravWaveNoticeSchema,
    _LocusSchema,
)
from .config import config
from .utils import mjd_to_datetime


class GravWaveNoticeTypes(Enum):
    EARLY_WARNING = "EARLY_WARNING"
    EARLYWARNING = "EARLY_WARNING"
    PRELIMINARY = "PRELIMINARY"
    INITIAL = "INITIAL"
    UPDATE = "UPDATE"
    RETRACTION = "RETRACTION"


@dataclass
class GravWaveNotice:
    gracedb_id: str
    notice_type: GravWaveNoticeTypes
    notice_datetime: datetime.datetime
    id: Optional[int] = field(default_factory=lambda: None)
    event_datetime: Optional[datetime.datetime] = field(default_factory=lambda: None)
    false_alarm_rate: Optional[float] = field(default_factory=lambda: None)
    skymap_base64: Optional[str] = field(default_factory=lambda: None)
    external_coinc: Optional[dict] = field(default_factory=lambda: None)
    full_notice: Optional[dict] = field(default_factory=lambda: None)
    version_id: int = field(default_factory=lambda: 1)

    _max_level: ClassVar[int] = 29
    _skymap: InitVar[Union[Table, None]] = None
    _healpix_nested_indices: InitVar[Union[np.ndarray, None]] = None
    _healpix_sorter: InitVar[Union[np.ndarray, None]] = None
    _probability_density_sorter: InitVar[Union[np.ndarray, None]] = None
    _sorted_pixel_areas: InitVar[Union[np.ndarray, None]] = None
    _cumprob: InitVar[Union[np.ndarray, None]] = None

    @property
    def skymap(self):
        if self._skymap is None:
            if self.skymap_base64 is None:
                return None
            skymap_bytes = b64decode(self.skymap_base64)
            self._skymap = Table.read(BytesIO(skymap_bytes))
        return self._skymap

    @property
    def healpix_nested_indices(self):
        if self._healpix_nested_indices is None:
            level, ipix = ah.uniq_to_level_ipix(self.skymap["UNIQ"])
            index = ipix * (2 ** (self._max_level - level)) ** 2
            self._healpix_nested_indices = index
        return self._healpix_nested_indices

    @property
    def healpix_sorter(self):
        # sorted indices by nested healpix indices
        if self._healpix_sorter is None:
            self._healpix_sorter = np.argsort(self.healpix_nested_indices)
        return self._healpix_sorter

    @property
    def probability_density_sorter(self):
        # sorted indices by probability density
        if self._probability_density_sorter is None:
            self._probability_density_sorter = np.argsort(self.skymap["PROBDENSITY"])
        return self._probability_density_sorter

    @property
    def sorted_pixel_areas(self):
        # pixel areas using probdensity sort order
        if self._sorted_pixel_areas is None:
            level, _ = ah.uniq_to_level_ipix(
                self.skymap["UNIQ"][self.probability_density_sorter]
            )
            pixel_area = ah.nside_to_pixel_area(ah.level_to_nside(level))
            self._sorted_pixel_areas = pixel_area
        return self._sorted_pixel_areas

    @property
    def cumprob(self):
        # cumulative probability using low-to-high prob density sort order
        if self._cumprob is None:
            prob = (
                self.sorted_pixel_areas
                * self.skymap["PROBDENSITY"][self.probability_density_sorter]
            )
            self._cumprob = np.cumsum(prob)  # low to high density
        return self._cumprob

    def get_probability_density(self, location: SkyCoord) -> float:
        max_nside = ah.level_to_nside(self._max_level)
        match_ipix = ah.lonlat_to_healpix(
            location.ra, location.dec, max_nside, order="nested"
        )

        i = self.healpix_sorter[
            np.searchsorted(
                self.healpix_nested_indices,
                match_ipix,
                side="right",
                sorter=self.healpix_sorter,
            )
            - 1
        ]
        return self.skymap[i]["PROBDENSITY"]

    def get_probability_contour_level_and_area(self, location: SkyCoord):
        prob_density = self.get_probability_density(location)

        i = (
            np.searchsorted(
                self.skymap["PROBDENSITY"],
                prob_density,
                side="right",
                sorter=self.probability_density_sorter,
            )
            - 1  # include duplicates
        )
        contour_level = (1.0 - self.cumprob[i].value) * 100.0
        area = self.sorted_pixel_areas[i:].sum().to(u.deg**2)
        return contour_level, area

    def to_devkit(self):
        return {
            "gracedb_id": self.gracedb_id,
            "notice_type": self.notice_type,
            "notice_datetime": self.notice_datetime,
            "id": self.id,
            "event_datetime": self.event_datetime,
            "false_alarm_rate": self.false_alarm_rate,
            "skymap_base64": self.skymap_base64,
            "external_coinc": self.external_coinc,
            "full_notice": self.full_notice,
            "version_id": self.version_id,
            "skymap": self.skymap,
        }


class AlertGravWaveEvent(TypedDict):
    gracedb_id: str
    contour_level: float
    contour_area: float


class Alert:
    """
    An ANTARES alert represents a single visit/observation of an astronomical object.

    Attributes
    ----------
    alert_id: str
        ANTARES ID for this alert.
    mjd: float
        Modified julian date of the alert.
    properties: dict
        Arbitrary, survey-specific properties associated with this alert.

    Note
    ----------
    processed_at and grav_wave_events are Optional to not break user code that
    uses the Alert class. This Optional doesn't apply to antares non client.
    """

    def __init__(
        self,
        alert_id: str,
        mjd: float,
        properties: dict,
        processed_at: Optional[datetime.datetime] = None,
        grav_wave_events: Optional[List[Optional[AlertGravWaveEvent]]] = None,
        **_,
    ):
        self.alert_id = alert_id
        self.mjd = mjd
        self.processed_at = processed_at
        self.properties = properties
        self.grav_wave_events = grav_wave_events


class Locus:
    """
    An ANTARES locus is a collection of metadata describing a single astronomical
    object.

    Attributes
    ----------
    locus_id: str
        ANTARES ID for this object.
    ra: float
        Right ascension of the centroid of alert history.
    dec: float
        Declination of the centroid of alert history.
    properties: dict
        A dictionary of ANTARES- and user-generated properties that are updated every
        time there is activity on this locus (e.g. a new alert).
    tags: List[str]
        A list of strings that are added to this locus by ANTARES- and user-submitted
        filters that run against the real-time alert stream.
    alerts: Optional[List[Alert]]
        A list of alerts that are associated with this locus. If `None`, the alerts
        will be loaded on first access from the ANTARES HTTP API.
    catalogs: Optional[List[str]]
        Names of catalogs that this locus has been associated with.
    catalog_objects: Optional[List[dict]]
        A list of catalog objects that are associated with this locus. If `None`, they
        will be loaded on first access from the ANTARES HTTP API.
    lightcurve: Optional[pd.DataFrame]
        Data frame representation of a subset of normalized alert properties. If `None`
        it will be loaded on first access from the ANTARES HTTP API.
    watch_list_ids: Optional[List[str]]
        A list of IDs corresponding to user-submitted regional watch lists.
    watch_object_ids: Optional[List[str]]
        A list of IDs corresponding to user-submitted regional watch list objects.
    grav_wave_events: Optional[List[str]]
        A list of gravitational wave event ids that are associated with this locus.

    Notes
    -----
    Instances of this class lazy-load a few of their attributes from the ANTARES API.
    These attributes are: `alerts`, `catalog_objects` and `lightcurve`.

    """

    def __init__(
        self,
        locus_id: str,
        ra: float,
        dec: float,
        properties: dict,
        tags: List[str],
        alerts: Optional[List[Alert]] = None,
        catalogs: Optional[List[str]] = None,
        catalog_objects: Optional[dict[str, list]] = None,
        lightcurve: Optional[pd.DataFrame] = None,
        watch_list_ids: Optional[List[str]] = None,
        watch_object_ids: Optional[List[str]] = None,
        grav_wave_events: Optional[List[str]] = None,
        **_,
    ):
        self.locus_id = locus_id
        self.ra = ra
        self.dec = dec
        self.properties = properties
        self.tags = tags
        self.catalogs = catalogs
        if self.catalogs is None:
            self.catalogs = []
        self.watch_list_ids = watch_list_ids
        if self.watch_list_ids is None:
            self.watch_list_ids = []
        self.watch_object_ids = watch_object_ids
        if self.watch_object_ids is None:
            self.watch_object_ids = []
        self.grav_wave_events = grav_wave_events
        if self.grav_wave_events is None:
            self.grav_wave_events = []
        self._alerts = alerts
        self._catalog_objects = catalog_objects
        self._lightcurve = lightcurve
        self._timeseries = None
        self._coordinates = None
        self._grav_wave_events_data = {}

    def _fetch_alerts(self) -> List[Alert]:
        alerts = _list_resources(
            config["ANTARES_API_BASE_URL"]
            + "/".join(("loci", self.locus_id, "alerts")),
            _AlertSchema,
        )
        return list(alerts)

    def _fetch_lightcurve(self) -> pd.DataFrame:
        locus = _get_resource(
            config["ANTARES_API_BASE_URL"] + "/".join(("loci", self.locus_id)),
            _LocusSchema,
        )
        return locus.lightcurve

    def _fetch_catalog_objects(self) -> dict:
        catalog_matches = _list_resources(
            config["ANTARES_API_BASE_URL"]
            + "/".join(("loci", self.locus_id, "catalog-matches")),
            _CatalogEntrySchema,
        )
        catalog_matches = list(catalog_matches)
        catalog_objects = defaultdict(list)
        for match in catalog_matches:
            catalog_name = match["catalog_entry_id"].split(":")[0]
            catalog_objects[catalog_name].append(match["properties"])
        return catalog_objects

    def _fetch_grav_wave_events_data(self) -> list[GravWaveNotice]:
        if self.grav_wave_events:
            return _list_resources(
                config["ANTARES_API_BASE_URL"]
                + "/grav_wave_notices/latest?ids={}".format(
                    ",".join(self.grav_wave_events)
                ),
                _GravWaveNoticeSchema,
            )
        else:
            return []

    @property
    def grav_wave_events_data(self) -> list[GravWaveNotice]:
        if self._grav_wave_events_data == {}:
            self._grav_wave_events_data = {
                notice.gracedb_id: notice
                for notice in self._fetch_grav_wave_events_data()
            }
        return self._grav_wave_events_data

    @grav_wave_events_data.setter
    def grav_wave_events_data(self, value) -> None:
        self._grav_wave_events_data = value

    @property
    def timeseries(self) -> astropy.timeseries.TimeSeries:
        """
        This `TimeSeries` contains all of the historical alert data associated with
        this object.
        """
        if self._timeseries is None:
            self._timeseries = astropy.timeseries.TimeSeries(
                data=[alert.properties for alert in self.alerts],
                time=[mjd_to_datetime(alert.mjd) for alert in self.alerts],
            )
        return self._timeseries

    @timeseries.setter
    def timeseries(self, value) -> None:
        self._timeseries = value

    @property
    def alerts(self) -> List[Alert]:
        """A list of alerts that are associated with this locus."""
        if self._alerts is None:
            self._alerts = self._fetch_alerts()
        return self._alerts

    @alerts.setter
    def alerts(self, value) -> None:
        self._alerts = value

    @property
    def catalog_objects(self) -> dict:
        """
        A dictionary of catalog objects that are associated with this locus. It has a
        structure like::

            {
                "<catalog_name">: [
                    { **<catalog_object_properties> },
                    { **<catalog_object_properties> },
                    ...
                ],
                ...
            }

        """
        if self._catalog_objects is None:
            self._catalog_objects = self._fetch_catalog_objects()
        return self._catalog_objects

    @catalog_objects.setter
    def catalog_objects(self, value) -> None:
        self._catalog_objects = value

    @property
    def lightcurve(self) -> pd.DataFrame:
        """Data frame representation of a subset of normalized alert properties."""
        if self._lightcurve is None:
            self._lightcurve = self._fetch_lightcurve()
        return self._lightcurve

    @lightcurve.setter
    def lightcurve(self, value) -> None:
        self._lightcurve = value

    @property
    def coordinates(self) -> SkyCoord:
        """Centroid of the locus as an AstroPy SkyCoord object."""
        if self._coordinates is None:
            self._coordinates = SkyCoord(f"{self.ra}d {self.dec}d")
        return self._coordinates

    def to_devkit(self, include_catalogs=False, include_grav_wave_events=False) -> dict:
        catalog_objects = {}
        if include_catalogs:
            catalog_objects = self.catalog_objects
        grav_wave_events_metadata = {}
        if include_grav_wave_events:
            grav_wave_events_metadata = {
                gracedb_id: notice.to_devkit()
                for gracedb_id, notice in self.grav_wave_events_data.items()
            }
        return {
            "id": self.locus_id,
            "ra": self.ra,
            "dec": self.dec,
            "user_tags": self.tags,
            "old_properties": self.properties,
            "catalog_objects": catalog_objects,
            "grav_wave_events_metadata": grav_wave_events_metadata,
            "alerts": [
                {
                    "id": alert.alert_id,
                    "mjd": alert.mjd,
                    "properties": alert.properties,
                    "grav_wave_events": alert.grav_wave_events,
                }
                for alert in self.alerts
            ],
        }



"""
Search the ANTARES database for objects of interest.
"""

import datetime
import itertools
import json
from collections import defaultdict
from typing import DefaultDict, Dict, Iterator, List, Optional
from urllib.parse import urljoin

import astropy.coordinates
import astropy.units

from ._api.api import (
    _get_raw_resource,
    _get_resource,
    _get_thumbnail_blob,
    _list_all_resources,
    _list_resources,
)
from ._api.schemas import (
    _AlertThumbnailSchema,
    _CatalogSampleSchema,
    _GravWaveNoticeSchema,
    _LocusListingSchema,
    _LocusSchema,
)
from .config import config
from .models import GravWaveNotice, Locus


def search(query: Dict) -> Iterator[Locus]:
    """
    Searches the ANTARES database for loci that meet certain criteria. Results are
    returned with the most recently updated objects first (sorted on the
    `properties.newest_alert_observation_time` field in descending order).

    Parameters
    ----------
    query: dict
        An ElasticSearch query. Must contain a top-level "query" key and only that
        top-level key. Other ES search arguments (e.g. "aggregations") are not allowed.

    Returns
    ----------
    Iterator over Locus objects

    """
    return _list_all_resources(
        urljoin(config["ANTARES_API_BASE_URL"], "loci"),
        _LocusListingSchema,
        params={
            "sort": "-properties.newest_alert_observation_time",
            "elasticsearch_query[locus_listing]": json.dumps(query),
        },
    )


def cone_search(
    center: astropy.coordinates.SkyCoord,
    radius: astropy.coordinates.Angle,
) -> Iterator[Locus]:
    """
    Searches the ANTARES database for loci in a certain region. Results are returned
    with the most recently updated objects first (sorted on the
    `properties.newest_alert_observation_time` field in descending order).

    Parameters
    ----------
    center: astropy.coordiantes.SkyCoord
    radius: astropy.coordiantes.Angle

    Returns
    ----------
    Iterator over Locus objects

    """
    return search(
        {
            "query": {
                "bool": {
                    "filter": {
                        "sky_distance": {
                            "distance": f"{radius.to_string(unit=astropy.units.deg, decimal=True)} degree",
                            "htm16": {"center": center.to_string()},
                        },
                    },
                },
            },
        }
    )


def get_by_id(locus_id: str) -> Optional[Locus]:
    """
    Gets an ANTARES locus by its ANTARES ID.

    Parameters
    ----------
    locus_id: str

    Returns
    ----------
    Locus or None

    """
    return _get_resource(
        urljoin(config["ANTARES_API_BASE_URL"], f"loci/{locus_id}"),
        _LocusSchema,
    )


def get_by_ztf_object_id(ztf_object_id: str) -> Optional[Locus]:
    """
    Gets an ANTARES locus by its ZTF Object ID.

    Parameters
    ----------
    ztf_object_id: str

    Returns
    ----------
    Locus or None

    """
    try:
        return next(
            search(
                {
                    "query": {
                        "bool": {
                            "filter": {
                                "term": {"properties.ztf_object_id": ztf_object_id},
                            },
                        },
                    },
                }
            )
        )
    except StopIteration:
        return None


# We can add get_by_alert_id and this can be done by adding an endpoint and using _get_locus_id_by_alert_id
def get_random_loci(n) -> list[Locus]:
    if n > 1000:
        raise ValueError("n must be less than or equal to 1000")
    # Build an ES query
    query = {
        "query": {
            "function_score": {
                "random_score": {},
            }
        }
    }
    return [locus for locus in itertools.islice(search(query), n)]


def get_random_locus() -> Locus:
    return get_random_loci(1)[0]


def get_random_locus_ids(n) -> list[str]:
    """
    Get `n` random locus_ids from the DB.
    """
    if n > 1000:
        raise ValueError("n must be less than or equal to 1000")
    return [locus.locus_id for locus in get_random_loci(n)]


def get_random_locus_id() -> str:
    """
    Get a random locus_id from the DB.
    """
    return get_random_locus_ids(1)[0]


def get_available_tags() -> List[str]:
    resource = _get_raw_resource(
        urljoin(config["ANTARES_API_BASE_URL"], "loci/statistics")
    )
    if resource.get("attributes") and resource.get("attributes").get("tags"):
        return list(resource.get("attributes").get("tags").keys())
    return []


def get_latest_grav_wave_notices(gracedb_id: str) -> GravWaveNotice:
    return _get_resource(
        urljoin(
            config["ANTARES_API_BASE_URL"],
            f"grav_wave_notices/{gracedb_id}/latest",
        ),
        _GravWaveNoticeSchema,
    )


def get_grav_wave_notices(
    gracedb_id: str, notice_datetime: datetime.datetime
) -> GravWaveNotice:
    return _get_resource(
        urljoin(
            config["ANTARES_API_BASE_URL"],
            f"grav_wave_notices/{gracedb_id}/{notice_datetime.isoformat()}",
        ),
        _GravWaveNoticeSchema,
    )


def get_multiple_grav_wave_notices(ids: List[str]) -> List[GravWaveNotice]:
    return list(
        _list_resources(
            urljoin(
                config["ANTARES_API_BASE_URL"],
                "grav_wave_notices/latest?ids={}".format(",".join(ids)),
            ),
            _GravWaveNoticeSchema,
        )
    )


def get_catalog_samples(n: int = 5) -> DefaultDict[str, List[Dict]]:
    catalog_samples = defaultdict(list)
    catalog_list = _list_resources(
        urljoin(config["ANTARES_API_BASE_URL"], f"catalog_samples?n={n}"),
        _CatalogSampleSchema,
    )
    for catalog_dict in catalog_list:
        catalog_name = catalog_dict["catalog_name"]
        catalog_samples[catalog_name].append(catalog_dict["data"])
    return catalog_samples


def catalog_search(ra: float, dec: float) -> DefaultDict[str, List[Dict]]:
    catalog_objects_combined = defaultdict(list)
    catalog_list = _list_resources(
        urljoin(config["ANTARES_API_BASE_URL"], f"catalog_search/{ra}/{dec}"),
        _CatalogSampleSchema,
    )
    for catalog_dict in catalog_list:
        catalog_name = catalog_dict["catalog_name"]
        catalog_objects_combined[catalog_name].append(catalog_dict["data"])
    return catalog_objects_combined


def get_thumbnails(alert_id: str):
    thumbnails = {}
    thumbnails_from_storage = _list_resources(
        urljoin(config["ANTARES_API_BASE_URL"], f"alerts/{alert_id}/thumbnails"),
        _AlertThumbnailSchema,
    )
    for thumbnail in thumbnails_from_storage:
        thumbnails[thumbnail["thumbnail_type"]] = {
            "alert_id": alert_id,
            "type": thumbnail["thumbnail_type"],
            "file_name": thumbnail["filename"],
            "blob": _get_thumbnail_blob(thumbnail["src"]),
        }
    return thumbnails



import marshmallow
from marshmallow import post_load
from marshmallow_jsonapi import Schema, fields

from .fields import _Lightcurve


class _ClientConfigStreamingSchema(Schema):
    class Meta:
        type_ = "client_config_streaming"
        unknown = marshmallow.EXCLUDE

    id = fields.String()
    type = fields.String()
    options = fields.Dict()


class _AlertSchema(Schema):
    class Meta:
        type_ = "alert"
        unknown = marshmallow.EXCLUDE

    id = fields.String(attribute="alert_id")
    properties = fields.Dict()
    processed_at = fields.DateTime()
    mjd = fields.Float()
    grav_wave_events = fields.List(fields.Dict())
    thumbnails = fields.Relationship()
    resource_meta = fields.ResourceMeta()
    document_meta = fields.DocumentMeta()

    @post_load
    def make_alert(self, data: dict, **_):
        from ..models import Alert

        return Alert(**data)


class _CatalogEntrySchema(Schema):
    class Meta:
        type_ = "catalog_entry"
        unknown = marshmallow.EXCLUDE

    id = fields.Str(attribute="catalog_entry_id")
    object_id = fields.Str()
    object_name = fields.Str()
    name = fields.Str()
    ra = fields.Float()
    dec = fields.Float()
    properties = fields.Dict()
    catalog = fields.Relationship()
    resource_meta = fields.ResourceMeta()
    document_meta = fields.DocumentMeta()


class _LocusSchema(Schema):
    class Meta:
        type_ = "locus"
        unknown = marshmallow.EXCLUDE

    id = fields.Str(attribute="locus_id")
    htm16 = fields.Int()
    ra = fields.Float()
    dec = fields.Float()
    grav_wave_events = fields.List(fields.Str())
    properties = fields.Dict()
    lightcurve = _Lightcurve()
    alerts = fields.Relationship()
    tags = fields.List(fields.Str())
    catalogs = fields.List(fields.Str())
    catalog_matches = fields.List(fields.Dict())
    resource_meta = fields.ResourceMeta()
    document_meta = fields.DocumentMeta()

    @post_load
    def make_locus(self, data: dict, **_):
        from ..models import Locus

        return Locus(**data)


class _LocusListingSchema(Schema):
    class Meta:
        type_ = "locus_listing"
        unknown = marshmallow.EXCLUDE

    id = fields.Str(attribute="locus_id")
    htm16 = fields.Int()
    ra = fields.Float()
    dec = fields.Float()
    properties = fields.Dict()
    locus = fields.Relationship()
    alerts = fields.Relationship()
    tags = fields.List(fields.Str())
    catalogs = fields.List(fields.Str())
    resource_meta = fields.ResourceMeta()
    document_meta = fields.DocumentMeta()

    @post_load
    def make_locus(self, data: dict, **_):
        from ..models import Locus

        return Locus(**data)


class _GravWaveNoticeSchema(Schema):
    class Meta:
        type_ = "grav_wave_notice"
        unknown = marshmallow.EXCLUDE

    id = fields.Int()
    gracedb_id = fields.Str()
    notice_type = fields.Str()
    notice_datetime = fields.DateTime(allow_none=True)
    event_datetime = fields.DateTime(allow_none=True)
    false_alarm_rate = fields.Float(allow_none=True)
    skymap_base64 = fields.Str(allow_none=True)
    external_coinc = fields.Dict(allow_none=True)
    full_notice = fields.Dict(allow_none=True)
    version_id = fields.Int(allow_none=True)

    @post_load
    def make_grav_wave_event(self, data: dict, **_):
        from ..models import GravWaveNotice, GravWaveNoticeTypes

        data["notice_type"] = GravWaveNoticeTypes(data["notice_type"])
        return GravWaveNotice(**data)


class _CatalogSampleSchema(Schema):
    class Meta:
        type_ = "catalog_sample"
        unknown = marshmallow.EXCLUDE

    id = fields.String(attribute="object_id")
    catalog_name = fields.String()
    data = fields.Dict()
    document_meta = fields.DocumentMeta()


class _AlertThumbnailSchema(Schema):
    class Meta:
        type_ = "alert_thumbnail"
        unknown = marshmallow.EXCLUDE

    id = fields.Str()
    filename = fields.Str()
    filemime = fields.Str()
    src = fields.Str()
    thumbnail_type = fields.Str()
    resource_meta = fields.Dict()
    document_meta = fields.DocumentMeta()



import abc
import itertools
import logging
import os
import re
import socket
import subprocess
import time
import zlib
from importlib.resources import files
from typing import Iterator, List, Optional, Tuple
from urllib.parse import urljoin

import bson
import confluent_kafka
from confluent_kafka.cimpl import (  # pylint: disable=no-name-in-module
    KafkaError,
    KafkaException,
)

from ._api.api import _get_resource
from ._api.schemas import _ClientConfigStreamingSchema, _LocusSchema
from .config import config
from .exceptions import AntaresException, IncompleteLocusException
from .models import Locus

log = logging.getLogger(__name__)

Topic = str


class AntaresNetworkingException(AntaresException):
    pass


class AntaresAlertParseException(AntaresException):
    pass


class AbstractStreamingClient(abc.ABC):
    @abc.abstractmethod
    def poll(
        self, timeout: Optional[float] = None
    ) -> Tuple[Tuple[None, None], Tuple[Topic, Locus]]:
        """
        Retrieve a single alert. This method blocks until ``timeout`` seconds have
        elapsed (by default, an infinite amount of time).

        Parameters
        ----------
        timeout: int
            Number of seconds to block waiting for an alert. If None, block indefinitely
            (default, None).

        Returns
        ----------
        (topic, locus): (str, Locus)
            Or ``(None, None)`` if ``timeout`` seconds elapse with no response
        """
        raise NotImplementedError

    def iter(self, limit: Optional[int] = None) -> Iterator[Tuple[Topic, Locus]]:
        """
        Yield from ANTARES alert streams.

        Parameters
        -----------
        limit: int
            Maximum number of messages to yield. If None, yield
            indefinitely (default, None).

        Yields
        ----------
        (topic, locus): str, Locus

        """
        for i in itertools.count(start=1, step=1):
            yield self.poll()
            if limit and i >= limit:
                return

    @abc.abstractmethod
    def commit(self):
        raise NotImplementedError

    @abc.abstractmethod
    def close(self):
        raise NotImplementedError

    @abc.abstractmethod
    def __enter__(self):
        raise NotImplementedError

    @abc.abstractmethod
    def __exit__(self, exc_type, exc_val, exc_tb):
        raise NotImplementedError


class KafkaStreamingClient(AbstractStreamingClient):
    _POLLING_FREQUENCY: float = 1.0
    _TOPIC_PREFIX = "client."
    _DEFAULT_SSL_CA_LOCATION: str = files(__package__).joinpath(
        "certificates/kafka-ca.pem"
    )

    def fetch_config(self):
        return _get_resource(
            urljoin(config["ANTARES_API_BASE_URL"], "client/config/streaming/default"),
            _ClientConfigStreamingSchema,
        )

    def __init__(self, topics: List[str], api_key: str, api_secret: str, **kwargs):
        """
        Creates a new ``KafkaStreamingClient`` instance.

        Parameters
        ----------
        topics: list of str
            Kafka stream topics to subscribe to.
        api_key: str
            API Key
        api_secret: str
            API Secret
        group: str, optional
            Group to connect to Kafka stream with. Changing this will reset
            your partition offsets. If you don't know what that means, DON'T
            pass any arguments for this (default, socket.gethostname()).
        ssl_ca_location: str, optional
            Path to your root SSL CAs cert.pem file.
        enable_auto_commit: bool, optional
            Enable automatic commits to the client's underlying Kafka streams
            (default, True).
        """
        self._topics = topics
        default_config = self.fetch_config()["options"]
        kafka_config = {
            "group.id": kwargs.get("group", socket.gethostname()),
            "logger": log,
            "default.topic.config": {"auto.offset.reset": "smallest"},
            "api.version.request": True,
            "broker.version.fallback": "0.10.0.0",
            "api.version.fallback.ms": 0,
            "enable.auto.commit": kwargs.get("enable_auto_commit", True),
            "sasl.username": api_key,
            "sasl.password": api_secret,
            "ssl.ca.location": kwargs.get("ssl_ca_location")
            or self._DEFAULT_SSL_CA_LOCATION,
            **default_config,
        }
        self._consumer = confluent_kafka.Consumer(kafka_config)
        self._consumer.subscribe(
            [f"{self._TOPIC_PREFIX}{topic}" for topic in self.topics]
        )

    def _timed_poll(self, timeout: float) -> Tuple[Optional[str], Optional[Locus]]:
        start_time = time.perf_counter()
        while (time.perf_counter() - start_time) < timeout:
            try:
                message = self._consumer.poll(timeout=self._POLLING_FREQUENCY)
                if message is not None:
                    locus = _parse_message(message)
                    return message.topic(), locus
            except KafkaException as kafka_exception:
                kafka_error = kafka_exception.args[0]
                # pylint: disable=protected-access
                if kafka_error == KafkaError._PARTITION_EOF:
                    pass
                # pylint: disable=protected-access
                elif kafka_error == KafkaError._TIMED_OUT:
                    exception_fmt = "There was an error connecting to ANTARES: {}"
                    raise AntaresNetworkingException(
                        exception_fmt.format(repr(kafka_exception))
                    ) from kafka_exception
                else:
                    exception_fmt = "There was an error consuming from ANTARES: {}"
                    raise AntaresException(
                        exception_fmt.format(repr(kafka_exception))
                    ) from kafka_exception
            except IncompleteLocusException:
                print("Skipping Incomplete Locus")
                return None, None
        return None, None

    def poll(self, timeout: float = None) -> Tuple[Optional[str], Optional[Locus]]:
        """
        Retrieve a single alert. This method blocks until ``timeout``
        seconds have elapsed (by default, an infinite amount of time).

        Parameters
        ----------
        timeout: int
            Number of seconds to block waiting for an alert. If None,
            block indefinitely (default, None).

        Returns
        ----------
        (topic, locus): (str, Locus)
            Or ``(None, None)`` if ``timeout`` seconds elapse with no response

        """
        if timeout:
            return self._timed_poll(timeout)
        locus = None
        while locus is None:
            topic, locus = self._timed_poll(self._POLLING_FREQUENCY)
        return topic, locus

    def commit(self):
        """Commit to the underlying Kafka stream."""
        self._consumer.commit()

    def close(self):
        """Close the client's connection."""
        self._consumer.close()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, exc_traceback):
        self._consumer.close()

    @property
    def topics(self):
        return self._topics


def _call(cmd):
    process = subprocess.Popen(
        cmd, shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE
    )
    stdout, stderr = process.communicate()
    return_code = process.returncode
    return return_code, stdout, stderr


def _merge_dictionaries(*dictionaries):
    merged_dictionary = {}
    for dictionary in dictionaries:
        merged_dictionary.update(
            {key: val for key, val in dictionary.items() if val is not None}
        )
    return merged_dictionary


def _locate_ssl_certs_file():
    """
    Attempt to locate openssl's CA certs file. Attempts to search
    a list of known locations first. Failing that, calls ``openssl``
    and tries to parse the output for file location.

    Raises
    ----------
    FileNotFoundError
      If no SSL certs file can be located.

    """
    # Check known locations first
    known_locations = [
        "/opt/local/etc/openssl/cert.pem",
        "/usr/local/etc/openssl/cert.pem",
        "/etc/pki/tls/cert.pem",
        "/etc/ssl/certs/ca-certificates.crt",
    ]
    log.info("Looking for openssl certs file in known locations.")
    for path in known_locations:
        log.debug("Checking location {}".format(path))
        if os.path.exists(path):
            log.info("Found certs at {}".format(path))
            return path
    # Failing that, try calling openssl directly
    log.info("Didn't find certs file in known locations.")
    log.info("Attempting to locate certs using `openssl version -d`")
    return_code, stdout, _ = _call("openssl version -d")
    if return_code != 0:
        log.info("openssl returned error code {}".format(return_code))
        log.error("Failed to locate openssl certs file.")
    else:
        regex = re.compile(r"OPENSSLDIR: \"(?P<path>.*)\"")
        log.debug("openssl stdout:")
        log.debug(stdout.decode())
        match = re.search(regex, stdout.decode())
        if match:
            path = os.path.join(match.group("path"), "cert.pem")
            if os.path.exists(path):
                log.info("Found certs at {}".format(path))
                return path
    # Failing that, raise an error
    raise FileNotFoundError("Could not locate SSL certificate")


def _parse_message(message):
    if message.error():
        raise KafkaException(message.error().code())
    locus_dict = bson.loads(zlib.decompress(message.value()))
    for key in ["ra", "dec", "properties", "tags"]:
        if key not in locus_dict["data"]["attributes"]:
            raise IncompleteLocusException(f"Missing field {key} on Locus")
    locus = _LocusSchema(partial=True).load(locus_dict)
    return locus


StreamingClient = KafkaStreamingClient



from typing import Any, Dict, Iterator, List, Optional, Type

import requests
from marshmallow_jsonapi import Schema
from typing_extensions import TypedDict

from antares_client.exceptions import AntaresException

from ..config import config

PageParam = TypedDict(
    "PageParam",
    {
        "limit": int,
        "offset": int,
    },
)

QueryParams = TypedDict(
    "QueryParams",
    {
        "sort": str,
        "page": PageParam,
        "fields": Dict[str, List[str]],
        "elasticsearch_query": Dict,
        # "filter": None  # RESERVED
    },
)


def _get_resource(
    url: str, schema_cls: Type[Schema], params: Optional[QueryParams] = None
) -> Optional[Any]:
    response = requests.get(url, params=params, timeout=config["API_TIMEOUT"])
    if response.status_code == 404:
        return None
    if response.status_code >= 400:
        raise AntaresException(response.json())
    return schema_cls(partial=True).load(response.json())


def _list_resources(
    url: str, schema_cls: Type[Schema], params: Optional[QueryParams] = None
) -> Iterator[Any]:
    response = requests.get(url, params=params, timeout=config["API_TIMEOUT"])
    if response.status_code >= 400:
        raise AntaresException(response.json())
    yield from schema_cls(many=True, partial=True).load(response.json())


def _list_all_resources(
    url: str, schema_cls: Type[Schema], params: Optional[QueryParams] = None
) -> Iterator[Any]:
    while True:
        response = requests.get(url, params=params, timeout=config["API_TIMEOUT"])
        if response.status_code >= 400:
            raise AntaresException(response.json())
        yield from schema_cls(many=True, partial=True).load(response.json())
        url = response.json().get("links", {}).get("next")
        if url is None:
            break
        params = None


def _get_raw_resource(url: str) -> Optional[Dict]:
    response = requests.get(url, timeout=config["API_TIMEOUT"])
    if response.status_code == 404:
        return None
    if response.status_code >= 400:
        raise AntaresException(response.json())
    data = response.json().get("data", {})
    return data


def _get_thumbnail_blob(url: str) -> Optional[bytes]:
    response = requests.get(url, timeout=config["API_TIMEOUT"])
    if not response.ok:
        print(f"Blob not found. Error: {response}")
        return None
    return response.content



from importlib.metadata import version

from . import search
from .exceptions import AntaresException

__version__ = version("antares_client")

try:
    from .stream import StreamingClient
except ImportError:

    class _MissingStreamingClient:
        def __init__(self, *args, **kwargs):
            raise ImportError(
                "To use KafkaStreamingClient you must install confluent_kafka or antares-client[subscriptions]"
            )

    StreamingClient = _MissingStreamingClient



from io import StringIO

import pandas as pd
from marshmallow import fields


class _Lightcurve(fields.Field):
    """Field that represents an ANTARES lightcurve as a pandas dataframe"""

    def _deserialize(self, value, attr, data, **kwargs):
        return pd.read_csv(StringIO(value))
